// 8.04.2019, Arduino IDE v1.8.9, LZ2WSG
//---------------------------------------------------------------------------------------------------------
#include "src/Rotary.h"              // http://www.buxtronix.net/2011/10/rotary-encoders-done-properly.html
#include "src/si5351.h"              // https://github.com/etherkit/Si5351Arduino, v2.1.0
#include "src/SSD1306AsciiAvrI2c.h"  // https://github.com/greiman/SSD1306Ascii

#define DEFAULT_RX   7015000
#define BAND_L       7000000
#define BAND_H       7200000
#define IF_FREQ       496500
#define OLED_I2C_ADDR   0x3C  // OLED I2C address
#define ENC_INC_1         50  // encoder increment 1
#define ENC_INC_2        500  // encoder increment 2
#define CW_TONE_HZ       700  // CW tone frequency (Hz)  
#define CW_DASH_LEN      3.0  // length of dash
#define BREAK_IN_DELAY     5  // break-in hang time in dots 

#define PIN_SM            A0  // S-meter voltage input
#define PIN_SWR           A1  // SWR-meter voltage input
#define PIN_KEYER_SPEED   A3  // speed potentiometer  
#define SW_STEP            4  // encoder step switch


#define PIN_PTT_REVE       5  // PTT reverse line output
#define PIN_KEY_LINE       6  // key line output
#define PIN_PTT_LINE       7  // PTT line output
#define PIN_TONE_OUT       8  // audio tone / 8 ohm speaker with 220 ohm resistor

#define PIN_TX_ON          9  // TX permission input
#define PIN_RX_TX         10  // manual key input
#define PIN_PADDLE_R      11  // paddle right input
#define PIN_PADDLE_L      12  // paddle left input

Rotary r = Rotary(2, 3);      // PIN2 and PIN3 are only pins checked by interrupt
Si5351 si5351(0x60);          // Si5351 I2C address
SSD1306AsciiAvrI2c oled;

uint32_t rx;
uint32_t enc_inc;
int8_t enc_dir = 0;           // -1 DIR_CCW, 0 DIR_NONE, 1 DIR_CCW
bool ch_flag = true;
uint16_t val_sm;
int val_speed; 
bool rx_tx;
bool rx_tx_old;
bool l_paddle;
bool r_paddle;


enum trx_state_e {
  E_STATE_RX, 
  E_STATE_TX
};  // state of the txcvr 

trx_state_e curr_state;

enum key_state_e {
  E_KEY_UP, 
  E_KEY_DOWN
};  // state of the key line 
key_state_e key_state;

int char_sent_ms;          // time of char ending
int curr_ms;               // current time
int dot_length_ms;         // keyer dot length 60ms dot=20wpm
int br_length_ms;          // breake-in length 
                         
//---------------------------------------------------------------------------------

void setup() {
  pinMode(2, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  pinMode(4, INPUT_PULLUP);
  
  pinMode(9, INPUT_PULLUP);
  pinMode(10, INPUT_PULLUP);
  pinMode(11, INPUT_PULLUP);
  pinMode(12, INPUT_PULLUP);

  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT); 
  pinMode(8, OUTPUT); 
  pinMode(13, OUTPUT);

  digitalWrite(PIN_KEY_LINE, HIGH); // set KEY line high
  digitalWrite(PIN_PTT_LINE, LOW);  // set PTT line low
  digitalWrite(PIN_PTT_REVE, HIGH); // set PTT reverse line high

  analogReference(INTERNAL);  // ref.voltage 5,0V or 1,1V (S-meter max voltage)

  attachInterrupt(0, rotary_encoder, CHANGE);
  attachInterrupt(1, rotary_encoder, CHANGE);

  si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 61000); // xtal 25MHz, corr. -610 Hz

  si5351.output_enable(SI5351_CLK0, 1);
  si5351.output_enable(SI5351_CLK1, 1);

  si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_8MA);
  si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_8MA);
  si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_8MA);

  si5351.set_freq(IF_FREQ * SI5351_FREQ_MULT, SI5351_CLK1);

  delay(500);
  oled.begin(&Adafruit128x32, OLED_I2C_ADDR);

  rx = DEFAULT_RX;

  rx_tx = digitalRead(PIN_RX_TX);
  rx_tx_old = !rx_tx;
  show_state();

  enc_inc = ENC_INC_1;
  show_inc();

  curr_state = E_STATE_RX;
  key_state = E_KEY_UP;

  si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2); 
  si5351.output_enable(SI5351_CLK2, 0);                   // short beep off

  oled.setCursor(2, 3);
  oled.setFont(labels);
  oled.print("1");                 // "1" is "RX" in labels.h
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

void loop() 
{
  if (ch_flag) {
    ch_flag = false;
    rx += enc_dir * enc_inc;
    enc_dir = 0;
    rx = constrain(rx, BAND_L, BAND_H);

    si5351.set_freq((rx + IF_FREQ) * SI5351_FREQ_MULT, SI5351_CLK0);

    show_freq(rx);
  }
  if (digitalRead(SW_STEP) == LOW) { // increment events
    if (button_state(SW_STEP) > 0) {
      set_inc();
      show_inc();
      do {
      }
      while (digitalRead(SW_STEP) == LOW);
    }
  }

if (digitalRead(PIN_TX_ON) == LOW) 
{rx_tx=true; show_state(); goto bypass1;}   //  Check if TX permission switch is ON 


//-- manual keyer code begin ------------------------------------------
    
  rx_tx = digitalRead(PIN_RX_TX);
  if (rx_tx != rx_tx_old)         // check if rx_tx state changed
  {
  char_sent_ms = millis(); 
  show_state();                   // if yes run RX/TX routine
  rx_tx_old = rx_tx;
  }
if (digitalRead(PIN_RX_TX) == LOW) goto bypass2;

//-- automatic keyer code begin ---------------------------------------

  val_speed = analogRead(PIN_KEYER_SPEED);
  //dot_length_ms = 100 + (val_speed-324)/6;      // 46ms....226ms
  dot_length_ms = 100 + (val_speed-424)/7;      // 40ms....185ms
  br_length_ms = dot_length_ms * BREAK_IN_DELAY;
  
  l_paddle = !digitalRead(PIN_PADDLE_L);  // !=inversion
  if(l_paddle){activate_state(E_STATE_TX); send_dot();} 

  r_paddle = !digitalRead(PIN_PADDLE_R);  // !=inversion
  if(r_paddle){ activate_state(E_STATE_TX); send_dash();} 

  curr_ms = millis();
 if((curr_state == E_STATE_TX) and (key_state == E_KEY_UP) 
 and (curr_ms - char_sent_ms) > br_length_ms)
  activate_state(E_STATE_RX);

//-------------------------------------------------------------------------

bypass1:

  val_sm = analogRead(PIN_SM);
  show_sm_old(); 
  goto bypass3;

bypass2:
  val_sm = analogRead(PIN_SWR);
  show_sm(); 

bypass3:  
;
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

void set_key_state(key_state_e k)
{
  switch (k)
  {
      case E_KEY_DOWN:
      {
        digitalWrite(PIN_KEY_LINE, LOW);
        digitalWrite(13, 1);  // for now, turn the Nano's LED on
        tone(PIN_TONE_OUT, CW_TONE_HZ);
        key_state = E_KEY_DOWN;
      }
      break;

      case E_KEY_UP:
      {
        digitalWrite(PIN_KEY_LINE, HIGH);
        digitalWrite(13, 0);  // for now, turn the Nano's LED off
        noTone(PIN_TONE_OUT);
        key_state = E_KEY_UP;
        char_sent_ms = millis();
      }
      break;
  }    
}

void activate_state(trx_state_e s)
{
  switch (s)
  {
      case E_STATE_RX:
      {
        if(curr_state == E_STATE_RX)
        {
          // already in receive state, nothing to do!
        }
        else
        { 
          digitalWrite(PIN_PTT_LINE, 0); 
	  digitalWrite(PIN_PTT_REVE, 1);
          oled.setCursor(2, 3);
          oled.setFont(labels);
          oled.print("1");                 // "1" is "RX" in labels.h
          curr_state = E_STATE_RX;
        }
      }
      break;

      case E_STATE_TX:
      {
        if(curr_state == E_STATE_TX)
        {
          // already in transmit state, nothing to do!
        }
        else
        {
          digitalWrite(PIN_PTT_LINE, 1);
	  digitalWrite(PIN_PTT_REVE, 0); 	 
          oled.setCursor(2, 3);
          oled.setFont(labels);
          oled.print("2");                 // "2" is "TX" in labels.h        
          curr_state = E_STATE_TX;
        }
      }
      break;
  }    
}

void send_dot()
{
  delay(1);
  si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 1);

  val_sm = analogRead(PIN_SWR); // ADC SWR range = 0...1024 (0...1V)
  if (val_sm > 1022)             // Adjust treshold level of Uswr
  { si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 0);
  set_key_state(E_KEY_UP);  } // Stop generation
 
  val_sm = analogRead(PIN_SM);
  val_sm = analogRead(PIN_SM);
  show_sm();
  set_key_state(E_KEY_DOWN);
 
  delay(dot_length_ms);  // 1dot

  si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 0);;
  set_key_state(E_KEY_UP); 
 
  delay(dot_length_ms - 10);  // pause length 10ms correction
}

void send_dash()
{ 
  delay(1);
  si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 1);

  val_sm = analogRead(PIN_SWR); // ADC SWR range = 0...1024 (0...1V)
  if (val_sm > 1022)             // Adjust treshold level of Uswr
  { si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 0);
  set_key_state(E_KEY_UP);  } // Stop generation

  val_sm = analogRead(PIN_SM);
  val_sm = analogRead(PIN_SM);
  show_sm();  
  set_key_state(E_KEY_DOWN); 

  delay(dot_length_ms * CW_DASH_LEN);  // 3dots
  
  si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 0); 
  set_key_state(E_KEY_UP); 

  delay(dot_length_ms - 10);  // pause length 10ms correction
}

//-------------------------------------------------------------------------
void rotary_encoder() {              // interrupt routine
  uint8_t result = r.process();

  if (result) {
    ch_flag = true;
    if (result == DIR_CW)
      enc_dir = 1;
    else
      enc_dir = -1;
  }
}
//--------------------------------------------------------------------------
void set_inc() {
  if (enc_inc == ENC_INC_1)
    enc_inc = ENC_INC_2;             // 500 Hz
  else
    enc_inc = ENC_INC_1;             // 50 Hz
}
//--------------------------------------------------------------------------
void show_inc() {
  oled.setCursor(109, 3);
  oled.setFont(labels);
  if (enc_inc == ENC_INC_1)
    oled.print("6");                 // 50 Hz
  else
    oled.print("9");                 // 500 Hz
}
//--------------------------------------------------------------------------
void show_freq(uint32_t sf_rx) {
  uint16_t fr;

  oled.setFont(lcdnums14x24);
  oled.setCursor(1, 0);
  fr = sf_rx / 1000000;
  if (fr < 10)
    oled.print(':');          // ':' is changed to ' ' in lcdnums14x24.h
  oled.print(fr);
  oled.print('.');
  fr = (sf_rx % 1000000) / 1000;
  if (fr < 100)
    oled.print('0');
  if (fr < 10)
    oled.print('0');
  oled.print(fr);
  oled.print('.');
  fr = (sf_rx % 1000) / 10;
  if (fr < 10)
    oled.print('0');
  oled.print(fr);
}
//---------------------------------------------------------------------------
void show_state() 
{
  oled.setCursor(2, 3);
  oled.setFont(labels);
  if (rx_tx)
    {//oled.print("1");                 // "1" is "RX" in labels.h
     si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
     si5351.output_enable(SI5351_CLK2, 0);
     set_key_state(E_KEY_UP);      
     }    
  
  else
  
    {oled.print("2");                 // "2" is "TX" in labels.h
     activate_state(E_STATE_TX);
     delay(1);
     si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
     si5351.output_enable(SI5351_CLK2, 1); 
     set_key_state(E_KEY_DOWN); 
     
     val_sm = analogRead(PIN_SWR); // ADC SWR range = 0...1024 (0...1V)
     if (val_sm > 1022)             // Adjust treshold level of Uswr
     { si5351.set_freq((rx-700) * SI5351_FREQ_MULT, SI5351_CLK2);
     si5351.output_enable(SI5351_CLK2, 0);
     set_key_state(E_KEY_UP);  } // Stop generation
     
     }
}

//---------------------------LINEAR SCALE-------------------------------
void show_sm_old() {
  
// 0...1V =1024 digits = 110/223/336/449/562/675/788/901/1014 digits
// in the mid of the scale 1 rectangle =2v (full scale is 20v)

  uint8_t a = 0;
  uint8_t m = 0;
  a = (val_sm + 3) / 113;   
  oled.setFont(pixels);
  oled.setCursor(24, 3);
  for (m = 0; m < a; m++)    // from m=0 to m=a with step m=1
    if (m < 6)
      oled.print('7');               // '5' - hollow rectangle, 6px
    else
      oled.print('8');               // '6' - filled rectangle, 6px
  for (m = a; m < 9; m++)
    oled.print('.');                 // '.' 1px
}
//-------------------------------------------------------------------------
//--------------------------QUADRATIC SCALE--------------------------------
void show_sm() {
  
// LINEAR input 0...1V =1024 digits = 100...1024 digits = 10db
// QUADRAIC INPUT S =0,6.5,7.0,7.5,8.0,8.5,9.0,+3,+6,+10 // 10x2db = 20db

  uint8_t a = 0;
  uint8_t m = 0;

  val_sm = sqrt(val_sm);
  val_sm = 34*val_sm;   // 32...34(adjust)*32=1024  // SQRT(32)=1024 
 
  a = (val_sm + 3) / 113;   

  oled.setFont(pixels);
  oled.setCursor(24, 3);
  for (m = 0; m < a; m++)    // from m=0 to m=a with step m=1
    if (m < 6)
      oled.print('7');               // '5' - hollow rectangle, 6px
    else
      oled.print('8');               // '6' - filled rectangle, 6px
  for (m = a; m < 9; m++)
    oled.print('.');                 // '.' 1px
}
//-------------------------------------------------------------------------
uint8_t button_state(uint8_t pin_bt) 
 {                    // push button state: 0(very fast), 1(fast), 2(slow)
  uint32_t t_start;
  uint32_t t_bt = 0;
  uint8_t bt_state;

  t_start = millis();
  while (!digitalRead(pin_bt)) {
    t_bt = millis() - t_start;
    if (t_bt > 1000)
      break;
  }
  if (t_bt < 20)                     // 'nothing' for debounce
    bt_state = 0;
  else if (t_bt < 400)               // fast
    bt_state = 1;
  else                               // slow
    bt_state = 2;
  return bt_state;
}

